import logging

from .grammar import Grammar, GrammarError, NonTerminal
from .lexer import Lexer
from .lr import LR

LOG = logging.getLogger(__name__)


class LR1(LR):

    def _closure(self, items):
        """ An item is a triple:
            (n, P, t) for P a Production,
                          0 <= n <= len(P.rhs),
                          t a terminal symbol
        """
        new = set(items)
        result = set()

        while len(new) > 0:
            result.update(new)
            for (n, p, a) in set(new):
                if n == len(p.rhs):
                    continue
                b = p.rhs[n]
                if not isinstance(b, NonTerminal):
                    continue
                for t in self.first(p.rhs[n + 1:] + (a,)):
                    new.update((0, r, t) for r in b.productions)
            new.difference_update(result)

        return frozenset(result)

    def _goto(self, items, symbol):
        """ goto(I, X)
            where I is a set of items and X is a grammar symbol,
            is the closure of the set of all items,
                A -> a X . b; t
            such that
                A -> a . X b; t
            is in I.
        """

        return self._closure({
            (n + 1, p, a)
            for (n, p, a) in items
            if n < len(p.rhs) and p.rhs[n] == symbol
        })

    def _sets_of_items(self, start):
        """ The starting production - which may be augmented - is passed in """

        c = set()
        symbols = self.symbols

        new = {self._closure({(0, start, Lexer.EOF)})}
        while len(new) > 0:
            c.update(new)
            for items in set(new):
                for x in symbols:
                    goto = self._goto(items, x)
                    if len(goto) > 0:
                        new.add(goto)
            new.difference_update(c)
        return c

    def compile(self, start, augment=True):
        self.nt._validate()
        self._clear_caches()

        if augment:
            start = self.augment(start)

        # We must have a single start production
        if len(start.productions) != 1:
            raise GrammarError(
                "There should be 1 production for {start}, not {n}"
                .format(start=start.id, n=len(start.productions))
            )

        start_ = list(start.productions)[0]

        self._start_symbol = start
        self._terminals = self.terminals

        self._compile(start_)
        return self._start_symbol

    def _compile(self, starting_production):
        c = self._sets_of_items(starting_production)
        sets = list(c)
        index = {s: n for n, s in enumerate(sets)}

        action = {}
        goto = {}
        s0 = None

        for i, items in enumerate(sets):
            for n, p, b in items:

                # if [A -> alpha . a beta; b] is in I_i,
                #  a a terminal, and goto(I_i, a) = I_j,
                #  then set action[i, a] to 'shift j'

                if n < len(p.rhs) and not isinstance(p.rhs[n], NonTerminal):
                    g = index[self._goto(items, p.rhs[n])]
                    a = p.rhs[n]

                    action[i, a] = self._conflict(action.get((i, a)),
                                                  LR.Shift(g),
                                                  lookahead=a)

                # if [A -> alpha . ; a] is in I_i, A != S',
                # then set action[i, a] to 'reduce A -> alpha'

                if n == len(p.rhs) and p != starting_production:
                    action[i, b] = self._conflict(action.get((i, b)),
                                                  LR.Reduce(p),
                                                  lookahead=b)

                # if [S' -> S . ; $] is in I_i,
                # then set action[i, $] to 'accept'

                if (n == len(p.rhs) and
                        p == starting_production and
                        b == Lexer.EOF):

                    action[i, b] = self._conflict(action.get((i, b)),
                                                  LR.Accept())

                # the initial state of the parser is the one constructed from
                # the set containing [S' -> . S; $]

                if n == 0 and p == starting_production and b == Lexer.EOF:
                    if s0 is not None:
                        raise GrammarError(
                            "Multiple start states found, {s0} and {i}"
                            .format(s0=s0, i=i))
                    s0 = i

            for nt in self.nonterminals:
                g = self._goto(items, nt)
                if g in index:
                    goto[i, nt] = index[g]

        if s0 is None:
            raise GrammarError("No start state found")

        self._action_table = action
        self._goto_table = goto
        self._s0 = s0

    def _clear_caches(self):
        super(LR1, self)._clear_caches()
        self._shift_reduce_conflicts = set()
        self._reduce_reduce_conflicts = set()

    def _conflict(self, current, new, lookahead=None):
        if current is None:
            return new

        if current == new:
            return current

        if isinstance(current, LR.Accept) or isinstance(new, LR.Accept):
            raise GrammarError("accept conflict found",
                               existing_action=current,
                               new_action=new,
                               )

        if isinstance(current, LR.Shift) and isinstance(new, LR.Shift):
            raise GrammarError("shift/shift conflict found",
                               existing_action=current,
                               new_action=new,
                               )

        if isinstance(current, LR.Shift) and isinstance(new, LR.Reduce):
            return self._conflict_shift_reduce(current, new, lookahead)

        if isinstance(current, LR.Reduce) and isinstance(new, LR.Shift):
            return self._conflict_shift_reduce(new, current, lookahead)

        if isinstance(current, LR.Reduce) and isinstance(new, LR.Reduce):
            return self._conflict_reduce_reduce(current, new)

        if isinstance(current, LR.Error):
            a = current.kwargs['a']
            b = current.kwargs['b']
            LOG.debug("extracting former actions from error:", a, b,
                      "new action is", new)
            if new == a or new == b:
                LOG.debug("... keeping error")
                return current
            result1 = self._conflict(a, new, lookahead)
            result2 = self._conflict(b, new, lookahead)
            LOG.debug("... bailing with new results:", result1, result2)

        raise GrammarError("unhandled conflict found:",
                           current=current,
                           new=new,
                           lookahead=lookahead)

    def _conflict_shift_reduce(self, shift, reduce, lookahead):
        try:
            level, assoc = self._terminal_precedence(lookahead)
            l2, a2 = self._rule_precedence(reduce.rule)
        except GrammarError:
            # Use a default rule.
            LOG.info("Defaulting to shift for s/r:", shift, reduce, lookahead)
            self._shift_reduce_conflicts.add((shift, reduce, lookahead))
            return shift

        if level > l2:
            return shift
        elif level < l2:
            return reduce
        else:
            if assoc is Grammar.Left:
                return reduce
            elif assoc is Grammar.Right:
                return shift
            else:
                return LR.Error("{} is non-associative".format(lookahead),
                                a=shift, b=reduce)

    def _conflict_reduce_reduce(self, r1, r2):
        if r1.rule.gen < r2.rule.gen:
            LOG.info("Preferring older rule for r/r:", r1, r2)
            self._shift_reduce_conflicts.add((r1, r2))
            return r1
        elif r1.rule.gen > r2.rule.gen:
            LOG.info("Preferring older rule for r/r:", r2, r1)
            self._shift_reduce_conflicts.add((r2, r1))
            return r2

        raise GrammarError("reduce/reduce conflict found", r1=r1, r2=r2)



from .grammar import NonTerminal, Production, GrammarError
from .lexer import Lexer


class GrammarOperations(object):
    def __init__(self, *args, **kwargs):
        super(GrammarOperations, self).__init__(*args, **kwargs)
        self._clear_caches()
        self._start_symbol = None

    def _clear_caches(self):
        self._first_cache = None
        self._follow_cache = None
        self._precedence_cache = None

    @property
    def start(self):
        return self._start_symbol

    def _first(self, symbol):
        if self._first_cache is not None:
            return self._first_cache[symbol]

        first = {t: {t} for t in self.terminals}
        for nt in self.nonterminals:
            for p in nt.productions:
                if len(p.rhs) == 0:
                    first[nt] = {Production.Empty}
                else:
                    first[nt] = set()

        changed = True

        while changed:
            changed = False
            for nt in self.nonterminals:
                for p in nt.productions:
                    for y in p.rhs:
                        empty = False
                        for a in first[y]:
                            if a == Production.Empty:
                                empty = True
                            else:
                                if a not in first[nt]:
                                    first[nt].add(a)
                                    changed = True
                        if not empty:
                            break
                    else:
                        if Production.Empty not in first[nt]:
                            first[nt].add(Production.Empty)
                            changed = True

        self._first_cache = first
        return first[symbol]

    def first(self, rhs):
        first = set()
        for symbol in rhs:
            if symbol == Lexer.EOF:
                first.add(Lexer.EOF)
                break
            elif Production.Empty in self._first(symbol):
                first.update(self._first(symbol)
                             .difference({Production.Empty}))
            else:
                first.update(self._first(symbol))
                break
        else:
            first.add(Production.Empty)

        return first

    def follow(self, nt, start=None):
        if self._follow_cache is not None:
            return self._follow_cache[nt]

        if start is None:
            start = self._start_symbol

        follow = {nt: set() for nt in self.nonterminals}
        follow[start].add(Lexer.EOF)

        changed = True

        while changed:
            changed = False
            for p in self.productions():
                for i, b in enumerate(p.rhs):
                    if not isinstance(b, NonTerminal):
                        continue
                    rest = p.rhs[i + 1:]
                    if len(rest) > 0:
                        additional = (self.first(rest)
                                      .difference({Production.Empty}))
                        if Production.Empty in self.first(rest):
                            additional.update(follow[p.nt])
                    else:
                        additional = set(follow[p.nt])

                    additional.difference_update(follow[b])
                    if len(additional) > 0:
                        follow[b].update(additional)
                        changed = True

        self._follow_cache = follow
        return follow[nt]

    def augment(self, start):
        # Augment the grammar
        wrap = NonTerminal('<start>', self)
        wrap(start,)(id)
        start = self.nt._nt['<start>'] = wrap
        start.gen = 0
        return start

    # Functions for managing terminal precedence

    def _terminal_precedence(self, term):
        """ Give a computed precedence and associativity
            for the supplied terminal symbol.

            Precedence levels are from 0 (lowest)
            and increase positively with tighter binding.
        """

        if self._precedence_cache is None:

            cache = {}
            for level, (assoc, terms) in enumerate(self._precedence):
                for t in terms:
                    if t in cache:
                        raise GrammarError(
                            "Terminal {t} has multiple precedences".
                            format(t=t))
                    cache[t] = (level, assoc)

            self._precedence_cache = cache

        try:
            return self._precedence_cache[term]
        except KeyError:
            raise GrammarError("No precedence given for terminal {t}"
                               .format(t=term))

    def _rule_precedence(self, prod):
        """ Give a computed precedence level for a production.
            This is the specified value if there is one,
            otherwise it'll be the precedence of
            its rightmost terminal symbol.
        """

        if prod.prec is not None:
            return self._terminal_precedence(prod.prec)

        rightmost = None
        for x in prod.rhs:
            if not isinstance(x, NonTerminal):
                rightmost = x

        try:
            return self._terminal_precedence(rightmost)
        except KeyError:
            raise GrammarError("No value given when looking for "
                               "precedence for rule {p}"
                               .format(p=prod))



from .grammar import Grammar, GrammarError, NonTerminal, Production
from .lr import LR
from .slr import SLR
from .lr1 import LR1
from .lexer import Lexer, LexerError

__all__ = (Grammar, GrammarError, NonTerminal, Production,
           LR, SLR, LR1,
           Lexer, LexerError)



class Grammar(object):

    def __init__(self, *args, **kwargs):
        super(Grammar, self).__init__(*args, **kwargs)
        self.nt = Nonterminals(self)
        self._precedence = []
        self._generation = 0

    def compile(self, start=None):
        raise NotImplementedError()

    def parse(self, stream=None):
        raise NotImplementedError()

    def productions(self, nt=None):
        return self.nt._productions(nt)

    @property
    def terminals(self):
        return self.nt._terminals()

    @property
    def nonterminals(self):
        return self.nt._nonterminals()

    @property
    def symbols(self):
        return self.terminals.union(self.nonterminals)

    # precedence markers
    Left = object()
    Right = object()
    NonAssoc = object()

    def precedence(self, assoc, *args):
        self._precedence.append((assoc, frozenset(args)))

    def left(self, *args):
        self.precedence(Grammar.Left, *args)

    def right(self, *args):
        self.precedence(Grammar.Right, *args)

    def nonassoc(self, *args):
        self.precedence(Grammar.NonAssoc, *args)

    # Support for rule ordering by age
    def _count(self):
        self._generation += 1
        return self._generation


class GrammarError(Exception):
    def __init__(self, *args, **kwargs):
        super(GrammarError, self).__init__(*args)
        self.kwargs = kwargs

    def __str__(self):
        s = super(GrammarError, self).__str__()
        if self.kwargs:
            s += (' {' +
                  ', '.join(k + ': ' + str(self.kwargs[k])
                            for k in sorted(self.kwargs)) +
                  '}')
        return s


class Nonterminals(object):

    def __init__(self, grammar):
        self._grammar = grammar
        self._nt = {}

    def __getattr__(self, id):
        if id.startswith('_'):
            return
        try:
            return self._nt[id]
        except KeyError:
            nt = NonTerminal(id, self._grammar)
            self._nt[id] = nt
            return nt

    def _validate(self):
        for id in self._nt:
            if len(self._nt[id].alts) == 0:
                raise GrammarError("Non-terminal {id} has no productions"
                                   .format(id=id))

    def _terminals(self):
        terminals = set()
        for id in self._nt:
            terminals.update(self._nt[id].terminals())
        return terminals

    def _nonterminals(self):
        return set(self._nt.values())

    def _productions(self, nt=None):
        if nt is not None:
            assert nt == self._nt[nt.id]
            return nt.productions

        prods = set()
        for id in self._nt:
            prods.update(self._nt[id].productions)
        return prods


class NonTerminal(object):

    def __init__(self, id, grammar):
        self.id = id
        self.grammar = grammar
        self.alts = []

    def __call__(self, *args, **kwargs):
        prec = kwargs.get('prec')

        def decorate(f):
            self.alts.append(
                Production(self, args, prec=prec, fn=f,
                           gen=self.grammar._count())
            )
        return decorate

    def terminals(self):
        terminals = set()
        for alt in self.alts:
            terminals.update(alt.terminals())
        return terminals

    @property
    def productions(self):
        return set(self.alts)

    def __str__(self):
        return self.id

    def __getitem__(self, args):
        if not isinstance(args, tuple):
            args = (args,)
        for p in self.alts:
            if p.rhs == args:
                return p
        raise KeyError(args)


class Production(object):

    Empty = object()

    def __init__(self, nt, rhs, prec=None, fn=None, gen=0):
        self.nt = nt
        self.rhs = rhs
        self.prec = prec
        self.fn = fn
        self.gen = gen

    def terminals(self):
        terminals = set()
        for symbol in self.rhs:
            if isinstance(symbol, NonTerminal):
                continue
            terminals.add(symbol)
        return terminals

    def __str__(self):
        s = "{id} -> {items}".format(
            id=self.nt,
            items=" ".join(str(s) for s in self.rhs)
        )
        if self.prec is not None:
            s += " %prec {}".format(self.prec)
        return s



import re


class Lexer(object):

    ActualNone = object()
    EOF = object()

    def __init__(self):
        self.rules = []

    def __call__(self, pattern, flags=re.MULTILINE | re.VERBOSE):
        pattern = re.compile(pattern, flags)

        def decorate(f):
            self.rules.append((pattern, f))
            return f
        return decorate

    def lex(self, input):
        """ Take a string and lex it """
        while input != "":
            for p, f in self.rules:
                result = p.match(input)
                if result:
                    break
            else:
                raise LexerError(input)

            input = input[result.end():]
            if len(p.groupindex) == 0:
                token = f(*result.groups())
            else:
                token = f(**result.groupdict())
            if token is not None:
                yield token if token is not Lexer.ActualNone else None
        yield Lexer.EOF


class LexerError(Exception):
    pass



import logging

from .grammar import Grammar, GrammarError
from .operations import GrammarOperations
from .lexer import Lexer

LOG = logging.getLogger(__name__)


class LR(GrammarOperations, Grammar):

    def term(self, term):
        if term in self._terminals:
            return term
        elif type(term) in self._terminals:
            return type(term)
        elif term == Lexer.EOF:
            return term
        raise GrammarError("Terminal {t} not in terminals"
                           .format(t=term))

    def action(self, state, term):
        return self._action_table.get((state, self.term(term)), LR.Error())

    def goto(self, state, nt):
        return self._goto_table.get((state, nt), LR.Error())

    def parse(self, stream):
        # The stack consists of (data, state) pairs
        stack = [(None, self._s0)]
        for a in stream:
            LOG.debug("read", a)
            while True:
                LOG.debug("stack =", stack)
                action = self.action(stack[-1][1], a)
                if isinstance(action, LR.Shift):
                    LOG.debug("  shift", a, "=>", action.state)
                    stack.append((a, action.state))
                    break
                if isinstance(action, LR.Reduce):
                    LOG.debug("  reduce", action.rule, "=>")
                    # We do it this way because lambda productions
                    # have a length of 0, and stack[:-0] doesn't do
                    # the right thing.
                    split = len(stack) - len(action.rule.rhs)
                    value = action.rule.fn(*[v for (v, s) in stack[split:]])
                    stack = stack[:split]
                    state = self.goto(stack[-1][1], action.nt)
                    LOG.debug("  => reduce:", state, "=", value)
                    stack.append((value, state))
                    continue
                if isinstance(action, LR.Accept):
                    LOG.debug("accept, stack =", stack)
                    return stack[-1][0]
                if isinstance(action, LR.Error):
                    raise GrammarError(action.message,
                                       stack=stack,
                                       token=a,
                                       **action.kwargs)

    class Shift(object):
        def __init__(self, state):
            self.state = state

        def __repr__(self):
            return "LR.Shift({s})".format(s=self.state)

        def __eq__(self, other):
            return isinstance(other, LR.Shift) and other.state == self.state

        def __ne__(self, other):
            return not(self == other)

        def __hash__(self):
            return hash(self.state)

    class Reduce(object):
        def __init__(self, rule):
            self.rule = rule

        def __repr__(self):
            return "LR.Reduce({r})".format(r=self.rule)

        def __eq__(self, other):
            return isinstance(other, LR.Reduce) and other.rule == self.rule

        def __ne__(self, other):
            return not(self == other)

        def __hash__(self):
            return hash(self.rule)

        @property
        def nt(self):
            return self.rule.nt

    class Accept(object):
        def __repr__(self):
            return "LR.Accept"

        def __eq__(self, other):
            return isinstance(other, LR.Accept)

        def __ne__(self, other):
            return not(self == other)

        def __hash__(self):
            return hash(LR.Accept)

    class Error(object):
        def __init__(self, message='', **kwargs):
            self.message = message
            self.kwargs = kwargs

        def __repr__(self):
            s = "LR.Error({})".format(self.message)
            if self.kwargs:
                s += (' {' +
                      ', '.join(k + ': ' + str(self.kwargs[k])
                                for k in sorted(self.kwargs)) +
                      '}')
            return s

        __str__ = __repr__



from .grammar import GrammarError, NonTerminal
from .lexer import Lexer
from .lr import LR


class SLR(LR):

    def _closure(self, items):
        """ An item is a pair:
            (n, P) for P a Production, 0 <= n <= len(P.rhs)
        """
        new = set(items)
        result = set()

        while len(new) > 0:
            result.update(new)
            for (n, p) in set(new):
                if n == len(p.rhs):
                    continue
                b = p.rhs[n]
                if isinstance(b, NonTerminal):
                    new.update((0, r) for r in b.productions)
            new.difference_update(result)

        return frozenset(result)

    def _goto(self, items, symbol):
        """ goto(I, X)
            where I is a set of items and X is a grammar symbol,
            is the closure of the set of all items,
                A -> a X . b
            such that
                A -> a . X b
            is in I.
        """

        return self._closure({
            (n + 1, p)
            for (n, p) in items
            if n < len(p.rhs) and p.rhs[n] == symbol
        })

    def _sets_of_items(self, start):
        """ The starting production - which may be augmented - is passed in """

        c = set()
        symbols = self.symbols

        new = {self._closure({(0, start)})}
        while len(new) > 0:
            c.update(new)
            for items in set(new):
                for x in symbols:
                    goto = self._goto(items, x)
                    if len(goto) > 0:
                        new.add(goto)
            new.difference_update(c)
        return c

    def compile(self, start, augment=True):
        self.nt._validate()
        self._clear_caches()

        if augment:
            start = self.augment(start)

        # We must have a single start production
        if len(start.productions) != 1:
            raise GrammarError(
                "There should be 1 production for {start}, not {n}"
                .format(start=start.id, n=len(start.productions))
            )

        start_ = list(start.productions)[0]

        self._start_symbol = start
        self._terminals = self.terminals

        self._compile(start_)
        return self._start_symbol

    def _compile(self, starting_production):
        c = self._sets_of_items(starting_production)
        sets = list(c)
        index = {s: n for n, s in enumerate(sets)}

        action = {}
        goto = {}
        s0 = None

        for i, items in enumerate(sets):
            for n, p in items:
                if n == 0 and p == starting_production:
                    if s0 is not None:
                        raise GrammarError(
                            "Multiple start states found, {s0} and {i}"
                            .format(s0=s0, i=i))
                    s0 = i
                if n < len(p.rhs) and not isinstance(p.rhs[n], NonTerminal):
                    g = self._goto(items, p.rhs[n])
                    if (i, p.rhs[n]) in action:
                        raise GrammarError(
                            "Conflicting actions for the production {p}"
                            .format(p=p))
                    action[i, p.rhs[n]] = LR.Shift(index[g])
                elif n == len(p.rhs):
                    if p == starting_production:
                        if (i, Lexer.EOF) in action:
                            raise GrammarError(
                                "Multiple actions for EOF")
                        action[i, Lexer.EOF] = LR.Accept()
                    else:
                        for a in self.follow(p.nt):
                            if (i, a) in action:
                                raise GrammarError(
                                    "Conflict in state {s} on terminal {t}"
                                    .format(s=i, t=a))
                            action[i, a] = LR.Reduce(p)

            for nt in self.nonterminals:
                g = self._goto(items, nt)
                if g in index:
                    goto[i, nt] = index[g]

        if s0 is None:
            raise GrammarError("No start state found")

        self._action_table = action
        self._goto_table = goto
        self._s0 = s0



import os.path
from setuptools import setup, find_packages

def read_file(fn):
    with open(os.path.join(os.path.dirname(__file__), fn)) as f:
        return f.read()

setup(
    name="yalr",
    version="0.0.5",
    description="Yet Another LR(1) implementation",
    long_description=read_file("README.md"),
    author="jan grant",
    author_email="jang@ioctl.org",
    url="https://github.com/jan-grant/python-yalr",
    license=read_file("LICENCE.md"),

    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: YACC",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Compilers",
        "Topic :: Software Development :: Interpreters",
    ],

    platforms=["any",],

    packages=find_packages(),

    install_requires=[
        "six",
    ],
)
